#include <FL/fl_ask.h>

#include "ImpressionistView.h"
#include "ImpressionistModel.h"

#define MENU_HEIGHT 22
#define MAX_BRUSH_SIZE 20

////////////////////////////////////////////////////////////////////////
// Class ImpressionistView Implementation
//
// This class is responsible for creating various windows and widgets
// and also for maintaining the consistency between the UI states and
// the model (document) states, e.g. slider values and check box 
// states.
////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
// Menu definition
////////////////////////////////////////////////////////////////////////

Fl_Menu_Item ImpressionistView::ms_mivMainMenu[] = {
	{ "&File",		0, 0, 0, FL_SUBMENU },
		{ "&Load Image...",          FL_ALT + 'L', (Fl_Callback *)ImpressionistView::LoadImage_cb },
		{ "Save Canvas &As...",      FL_ALT + 'A', (Fl_Callback *)ImpressionistView::SaveImage_cb, 0, FL_MENU_DIVIDER },
		{ "E&xit",                   FL_ALT + 'X', (Fl_Callback *)ImpressionistView::Exit_cb },
		{ 0 },

	{ "&Tools", 0, 0, 0, FL_SUBMENU },
		{ "&Paint Toolbox",          FL_ALT + 'P', (Fl_Callback *)ImpressionistView::PaintToolbox_cb },
		{ 0 },
		 
	{ "&View", 0, 0, 0, FL_SUBMENU },
		{ "&Single Buffer",  FL_ALT + 'S', (Fl_Callback *)ImpressionistView::SingleBuffer_cb, 0, FL_MENU_RADIO | FL_MENU_VALUE },
		{ "&Double Buffer",  FL_ALT + 'D', (Fl_Callback *)ImpressionistView::DoubleBuffer_cb, 0, FL_MENU_RADIO },
		{ 0 },
		 
	{ "&Help",		0, 0, 0, FL_SUBMENU }, 
		{ "&About",	FL_ALT + 'A', (Fl_Callback *)ImpressionistView::About_cb },
		{ 0 },

	{ 0 }
};

Fl_Menu_Item ImpressionistView::ms_mivBrushTypeMenu[] = {
	{ "&Points",           FL_ALT+'P', (Fl_Callback*)ImpressionistView::BrushChoice_cb, (void*)BRUSH_POINT },
	{ "&Lines",            FL_ALT+'L', (Fl_Callback*)ImpressionistView::BrushChoice_cb, (void*)BRUSH_LINE },
	{ "&Circles",          FL_ALT+'C', (Fl_Callback*)ImpressionistView::BrushChoice_cb, (void*)BRUSH_CIRCLE },
	{ "Scattered Points",  FL_ALT+'Q', (Fl_Callback*)ImpressionistView::BrushChoice_cb, (void*)BRUSH_SCATTERED_POINTS },
	{ "Scattered Lines",   FL_ALT+'M', (Fl_Callback*)ImpressionistView::BrushChoice_cb, (void*)BRUSH_SCATTERED_LINES },
	{ "Scattered Circles", FL_ALT+'D', (Fl_Callback*)ImpressionistView::BrushChoice_cb, (void*)BRUSH_SCATTERED_CIRCLES },
	{ 0 }
}; 

////////////////////////////////////////////////////////////////////////
// Help functions
////////////////////////////////////////////////////////////////////////

// From menu item back to view itself
ImpressionistView* ImpressionistView::WhoAmI(Fl_Menu_* o)
{
	return ((ImpressionistView*)(o->parent()->user_data()));
}

////////////////////////////////////////////////////////////////////////
// Callback functions
////////////////////////////////////////////////////////////////////////

void ImpressionistView::LoadImage_cb(Fl_Menu_* o, void* v) 
{
	ImpressionistView* pvView = WhoAmI(o);
	ImpressionistModel* pmModel = WhoAmI(o)->m_pmModel;

	char* szNewFileName = fl_file_chooser("Load Image", "*.bmp", NULL);
	if (szNewFileName != NULL) {
		if (pmModel->LoadImage(szNewFileName)) {
			pvView->ResizeMainWnd(pmModel->ImageWidth() * 2, pmModel->ImageHeight() + MENU_HEIGHT);

			// display it in original view
			pvView->m_pwndOrigView->resize_window(pmModel->ImageWidth(), pmModel->ImageHeight());
			pvView->m_pwndOrigView->redraw();

			// refresh paint view as well
			pvView->m_pwndPaintView->resize_window(pmModel->ImageWidth(), pmModel->ImageHeight());
			pvView->m_pwndPaintView->redraw();

			pvView->UpdateWidgets();
		}
		else {
			fl_alert("It must be a great picture, but sorry, I can't load the bitmap file.");
		}
	}
}

void ImpressionistView::SaveImage_cb(Fl_Menu_* o, void* v) 
{
	ImpressionistModel *pmModel = WhoAmI(o)->m_pmModel;

	char* szNewFileName = fl_file_chooser("Save Canvas As", "*.bmp", NULL);
	if (szNewFileName != NULL) {
		if (pmModel->SaveImage(szNewFileName)) {
		}
		else {
			fl_alert("Sorry, I can't save your great painting to the bitmap file.");
		}
	}
}

void ImpressionistView::PaintToolbox_cb(Fl_Menu_* o, void* v) 
{
	WhoAmI(o)->m_pwndPaintDlg->show();
}

void ImpressionistView::SingleBuffer_cb(Fl_Menu_* o, void* v)
{
	ImpressionistView* pvView = WhoAmI(o);
	// faster, but you might lose your paint job
	pvView->m_pwndPaintView->mode(FL_RGB | FL_SINGLE);
}

void ImpressionistView::DoubleBuffer_cb(Fl_Menu_* o, void* v)
{
	ImpressionistView* pvView = WhoAmI(o);
	// a lot slower on some machines, but perhaps safer
	pvView->m_pwndPaintView->mode(FL_RGB | FL_DOUBLE);
}

void ImpressionistView::Exit_cb(Fl_Menu_* o, void* v) 
{
	WhoAmI(o)->m_pwndMainWnd->hide();
	WhoAmI(o)->m_pwndPaintDlg->hide();
}

void ImpressionistView::Exit2_cb(Fl_Widget* o, void* v) 
{
	ImpressionistView* pvView = (ImpressionistView*)(o->user_data());
	pvView->m_pwndMainWnd->hide();
	pvView->m_pwndPaintDlg->hide();
}

void ImpressionistView::About_cb(Fl_Menu_* o, void* v) 
{
	fl_message("CSE591 Project 1 - Impressionist");
}

void ImpressionistView::BrushChoice_cb(Fl_Widget* o, void* v)
{
	ImpressionistView* pvView = (ImpressionistView*)(o->user_data());
	ImpressionistModel* pmModel = pvView->m_pmModel;

	pmModel->BrushType((int)v);
	pvView->UpdateWidgets();
}

void ImpressionistView::SizeSlider_cb(Fl_Widget* o, void* v)
{
	((ImpressionistView*)(o->user_data()))->m_pmModel->BrushSize((int)((Fl_Slider*)o)->value());
}

void ImpressionistView::ClearButton_cb(Fl_Widget* o, void* v)
{
	ImpressionistView* pvView = (ImpressionistView*)(o->user_data());
	ImpressionistModel* pmModel = pvView->m_pmModel;

	pmModel->ClearCanvas();
	pvView->m_pwndPaintView->redraw();
}

////////////////////////////////////////////////////////////////////////
// Per instance functions
////////////////////////////////////////////////////////////////////////
 
void ImpressionistView::Show() 
{
	m_pwndMainWnd->show();
	m_pwndPaintView->show();
	m_pwndOrigView->show();
}

ImpressionistView::ImpressionistView(ImpressionistModel* pmModel)
{
	m_pmModel = pmModel;

	m_pwndMainWnd = new Fl_Window(600, 300, "Impressionist");
		m_pwndMainWnd->user_data((void*)(this));	
		// install menu bar
		m_pmbMenuBar = new Fl_Menu_Bar(0, 0, 600, MENU_HEIGHT);
		m_pmbMenuBar->menu(ms_mivMainMenu);
		m_pmbMenuBar->textsize(11);

		Fl_Group* group = new Fl_Group(0, MENU_HEIGHT, 600, 300 - MENU_HEIGHT);

			// install original view window
			m_pwndOrigView = new OriginalView(0, MENU_HEIGHT, 300, 300 - MENU_HEIGHT, "Original View");
			m_pwndOrigView->deactivate();
			m_pwndOrigView->set_model(pmModel);

			// install paint view window
			m_pwndPaintView = new PaintView(300, MENU_HEIGHT, 300, 300 - MENU_HEIGHT, "Paint View");
			m_pwndPaintView->mode(FL_RGB | FL_SINGLE);
			m_pwndPaintView->set_model(pmModel);

		group->end();
		Fl_Group::current()->resizable(group);

		m_pwndMainWnd->callback(Exit2_cb);	
		m_pwndMainWnd->when(FL_HIDE);
    m_pwndMainWnd->end();

	int y;

	// paint dialog
	m_pwndPaintDlg = new Fl_Window(310, 90, "Paint");
		// group box
		m_pbxManualPaintBox = new Fl_Box(5, 5, 300, 80, "");
		m_pbxManualPaintBox->box(FL_ENGRAVED_BOX);

		y = 10;
		m_pchoBrushTypeChoice = new Fl_Choice(45, y, 150, 20, "&Brush");
		m_pchoBrushTypeChoice->user_data((void*)(this)); 
        m_pchoBrushTypeChoice->labelsize(11);
        m_pchoBrushTypeChoice->textsize(11);
		m_pchoBrushTypeChoice->menu(ms_mivBrushTypeMenu);
		m_pchoBrushTypeChoice->callback(BrushChoice_cb);

		y += 25;
		m_pslBrushSizeSlider = new Fl_Value_Slider(10, y, 180, 20, "Brush Size");
		m_pslBrushSizeSlider->user_data((void*)(this)); 
		m_pslBrushSizeSlider->type(FL_HOR_NICE_SLIDER);
        m_pslBrushSizeSlider->labelfont(FL_HELVETICA);
        m_pslBrushSizeSlider->labelsize(11);
		m_pslBrushSizeSlider->minimum(0);
		m_pslBrushSizeSlider->maximum(MAX_BRUSH_SIZE);
		m_pslBrushSizeSlider->step(1);
		m_pslBrushSizeSlider->align(FL_ALIGN_RIGHT);
		m_pslBrushSizeSlider->callback(SizeSlider_cb);

		y += 25;
		m_btClear = new Fl_Button(10, y, 60, 20, "Clear");
		m_btClear->user_data((void*)this);
        m_btClear->labelsize(11);
		m_btClear->callback(ClearButton_cb);
		m_btClear->when(FL_WHEN_RELEASE);

		// TODO: Add other widgets as needed
    m_pwndPaintDlg->end();	

	UpdateWidgets();
}

ImpressionistView::~ImpressionistView()
{
	// FLTK will delete other child windows for us
	delete m_pwndMainWnd;
	delete m_pwndPaintDlg;
}

void ImpressionistView::UpdateWidgets()
{
	// Update widget values based on the current model states
	m_pchoBrushTypeChoice->value(m_pmModel->BrushType());
	m_pslBrushSizeSlider->value(m_pmModel->BrushSize());
}

void ImpressionistView::ResizeMainWnd(int iWidth, int iHeight) 
{
	if (m_pwndMainWnd) {
		m_pwndMainWnd->resize(m_pwndMainWnd->x(), m_pwndMainWnd->y(), iWidth, iHeight);
	}
}

