#include <math.h>
#ifdef _DEBUG
#include <assert.h>
#endif // _DEBUG
#include "ImpressionistModel.h"
#include "ImpressionistView.h"
#include "PaintView.h"
#include "OriginalView.h"
#include "Brush.h"
#include "DotBrush.h"
#include "Color.h"

#define LEFT_MOUSE_DOWN		1
#define LEFT_MOUSE_DRAG		2
#define LEFT_MOUSE_UP		3
#define RIGHT_MOUSE_DOWN	4
#define RIGHT_MOUSE_DRAG	5
#define RIGHT_MOUSE_UP		6

#ifndef min
#define min(a, b) ((a) < (b) ? (a) : (b))
#endif

////////////////////////////////////////////////////////////////////////
// PaintView Class Implementation
//
// This window is resposible for displaying the painted image (canvas)
// and also the interaction between the user and the canvas.
// 
// If you plan to implement the stylized painting algorithm, this is
// a good place to handle it (as well as in ImpressionistModel) with
// the brushes you implement.
////////////////////////////////////////////////////////////////////////

PaintView::PaintView(int x, int y, int w, int h, const char *l) : 
	Fl_Gl_Window(x, y, w, h, l)
{
	// create one instance of each brush
	m_ppbrBrushes = new Brush*[BRUSH_COUNT];
	m_ppbrBrushes[BRUSH_POINT] = new DotBrush();
	// TODO: Replace these stinky DotBrushes with your fancy new brushes!
	m_ppbrBrushes[BRUSH_LINE] = new DotBrush();
	m_ppbrBrushes[BRUSH_CIRCLE] = new DotBrush();
	m_ppbrBrushes[BRUSH_SCATTERED_POINTS] = new DotBrush();
	m_ppbrBrushes[BRUSH_SCATTERED_LINES] = new DotBrush();
	m_ppbrBrushes[BRUSH_SCATTERED_CIRCLES] = new DotBrush();

	m_nWindowWidth = w;
	m_nWindowHeight = h;
	m_bLButtonDown = false;
	m_bRButtonDown = false;

	m_bIsAnEvent = false;
	m_pbyCanvas = NULL;
}

PaintView::~PaintView() 
{
	if (m_pbyCanvas) 
		delete[] m_pbyCanvas;

	for (int iBrush = 0; iBrush < BRUSH_COUNT; ++iBrush) {
		delete m_ppbrBrushes[iBrush];
	}
}

void PaintView::draw()
{
	Color cColor;
	bool bNothingDrawn = true;

	// Do one-time initialization
	if (!valid()) {
		glDrawBuffer(GL_BACK);

		m_nWindowWidth = w();
		m_nWindowHeight = h();
		m_nDrawWidth = min(m_nWindowWidth, m_pmModel->ImageWidth());
		m_nDrawHeight = min(m_nWindowHeight, m_pmModel->ImageHeight());

		glViewport(0, 0, m_nWindowWidth, m_nWindowHeight);

		glClearColor(1.0f, 1.0f, 1.0f, 1.0); // white
		glShadeModel(GL_FLAT);
		glDisable(GL_DEPTH_TEST);

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		glLoadIdentity();
		glViewport(0, m_nWindowHeight - m_nDrawHeight, m_nDrawWidth, m_nDrawHeight);
		glOrtho(0, m_nDrawWidth, 0, m_nDrawHeight, -1, 1);
	}

	// Image not loaded yet. Just show a blank canvas
	if (!m_pmModel->HasBitmap()) {
		glClear(GL_COLOR_BUFFER_BIT);
		return;
	}

	m_nStartRow = m_pmModel->ImageHeight() - m_nDrawHeight;
	if (m_nStartRow < 0) 
		m_nStartRow = 0;
	m_nEndRow = m_nStartRow + m_nDrawHeight;
	m_nStartCol = 0;
	m_nEndCol = m_nStartCol + m_nDrawWidth;

#ifdef _DEBUG
	// The canvas buffer should be allocated by now
	assert(m_pbyCanvas);
#endif // _DEBUG
	m_pPaintBitstart = m_pbyCanvas + 
		3 * m_pmModel->ImageWidth() * m_nStartRow;

	if (m_bIsAnEvent) {
		m_bIsAnEvent = false;

		Point ptSource(m_ptCoord.x + m_nStartCol, m_nEndRow - m_ptCoord.y);
		Point ptTarget(m_ptCoord.x, m_nDrawHeight - m_ptCoord.y);
		// Get the color of this stroke
		const GLubyte* pbySource = m_pmModel->GetOriginalPixel(ptSource);
		cColor.r = pbySource[0];
		cColor.g = pbySource[1];
		cColor.b = pbySource[2];
		cColor.a = 255;

		Brush* pbrBrush = m_ppbrBrushes[m_pmModel->BrushType()];

		switch (m_iEventToDo) {
		case LEFT_MOUSE_DOWN:
			pbrBrush->BrushBegin(m_pmModel->BrushSize(), cColor, ptTarget);
			bNothingDrawn = false;
			break;

		case LEFT_MOUSE_DRAG:
			pbrBrush->BrushMove(m_pmModel->BrushSize(), cColor, ptTarget);
			bNothingDrawn = false;
			break;

		case LEFT_MOUSE_UP:
			pbrBrush->BrushEnd(m_pmModel->BrushSize(), cColor, ptTarget);
			save_current_content();
			bNothingDrawn = false;
			break;

		case RIGHT_MOUSE_DOWN:
			break;

		case RIGHT_MOUSE_DRAG:
			break;

		case RIGHT_MOUSE_UP:
			break;

		default:
			break;
		}
	}

	if (bNothingDrawn) {
		// Probably a "repaint" message. Just paint the window 
		// with our saved content.
		restore_content();
	}
}

int PaintView::handle(int event)
{
	switch (event) {
	case FL_MOVE:
		m_ptCoord.x = Fl::event_x();
		m_ptCoord.y = Fl::event_y();
		break;

	case FL_PUSH:
		m_ptCoord.x = Fl::event_x();
		m_ptCoord.y = Fl::event_y();
		if (Fl::event_button() > 1) {
			m_bRButtonDown = true;
			m_iEventToDo = RIGHT_MOUSE_DOWN;
		}
		else {
			m_bLButtonDown = true;
			m_iEventToDo = LEFT_MOUSE_DOWN;
		}
		m_bIsAnEvent = true;
		redraw();
		break;

	case FL_DRAG:
		m_ptCoord.x = Fl::event_x();
		m_ptCoord.y = Fl::event_y();
		if (Fl::event_button() > 1)
			if (m_bRButtonDown)
				m_iEventToDo = RIGHT_MOUSE_DRAG;
			else {
				m_bRButtonDown = true;
				m_iEventToDo = RIGHT_MOUSE_DOWN;
			}
		else
			if (m_bLButtonDown)
				m_iEventToDo = LEFT_MOUSE_DRAG;
			else {
				m_bLButtonDown = true;
				m_iEventToDo = LEFT_MOUSE_DOWN;
			}
		m_bIsAnEvent = true;
		redraw();
		break;

	case FL_RELEASE:
		m_ptCoord.x = Fl::event_x();
		m_ptCoord.y = Fl::event_y();
		if (Fl::event_button() > 1) {
			m_bRButtonDown = false;
			m_iEventToDo = RIGHT_MOUSE_UP;
		}
		else {
			m_bLButtonDown = false;
			m_iEventToDo = LEFT_MOUSE_UP;
		}
		m_bIsAnEvent = true;
		redraw();
		break;

	default:
		break;
	}

	return 1;
}

void PaintView::resize_window(int iWidth, int iHeight)
{
	resize(x(), y(), iWidth, iHeight);
	// allocate buffer for canvas
	if (m_pbyCanvas) {
		m_pmModel->Canvas(NULL);
		delete[] m_pbyCanvas;
	}
	m_pbyCanvas = new unsigned char[iWidth * iHeight * 3];
	m_pmModel->Canvas(m_pbyCanvas);
	// set the canvas to all white
	::memset(m_pbyCanvas, 255, iWidth * iHeight * 3);
}

void PaintView::save_current_content()
{
	glPixelStorei(GL_PACK_ALIGNMENT, 1);
	glPixelStorei(GL_PACK_ROW_LENGTH, m_pmModel->ImageWidth());
	glReadPixels(0, m_nWindowHeight - m_nDrawHeight, 
		m_nDrawWidth, m_nDrawHeight, 
		GL_RGB, GL_UNSIGNED_BYTE, 
		m_pPaintBitstart);
}

void PaintView::restore_content()
{
	glClear(GL_DEPTH_BUFFER_BIT);

	glRasterPos2i(0, 0);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, m_pmModel->ImageWidth());
	glDisable(GL_BLEND);
	glDrawPixels(m_nDrawWidth, m_nDrawHeight, 
		GL_RGB, GL_UNSIGNED_BYTE, 
		m_pPaintBitstart);
	glEnable(GL_BLEND);
}
